#if !defined HATP_LIGHT_ENTITY_HH_
#define HATP_LIGHT_ENTITY_HH_

#include<iostream>
#include<string>
#include<list>
#include<cassert>
#include<typeinfo>
#include<boost/lexical_cast.hpp>

#include "Entity.hh"
#include "AttributeSet.hh"
#include "LightAttributeSet.hh"

/** Represents any entity in the HATP domains.
 */
class LightEntity {

protected:
	/** Entity's name */
	std::string name;
	/** Entity's type */
	std::string type;

	/**List of number attributes.
	* The implementation of a number is a double.*/
	LightAttributeSet<number> numberAttrs;
	/**List of string attributes*/
	LightAttributeSet<std::string> stringAttrs;
	/**List of bool attributes*/
	LightAttributeSet<bool> boolAttrs;
	/**List of Entity attributes*/
	AttributeSet<LightEntity*> entityAttrs;

	/** Store the fact that the Entity is unknown for the agent that owns it.
	*/
	bool unknown;
	/** Store the fact that the Entity is known for the agent that owns it.
	*/
	bool known;

	/** Set of unknown attributes.
	* Unkown is used in some domain to say that the Entity's attributes exists but is unkown by an agent.
	*/
	std::set<std::string> unknownAttrs;

	/** Set of known attributes.
	* Known is used in some domain to say that the Entity's attributes exists and is kown by an agent.
	*/
	std::set<std::string> knownAttrs;

public:
	LightEntity();
	LightEntity(std::string n);
	explicit LightEntity(Entity const& e);
	LightEntity(LightEntity const& e);
	std::string const& getName() const;
	std::string const& getType() const;
	bool isUnknown() const;
	bool isKnown() const;
	bool isAttrUnknown(const std::string& attrName) const;
	bool isAttrKnown(const std::string& attrName) const;
	std::set<std::string> const& getUnknownAttrs() const;
	std::set<std::string> const& getKnownAttrs() const;
	void setUnknown(bool f);
	void setKnown(bool f);
	void setAttrUnknown(const std::string& attrName, bool f);
	void setAttrKnown(const std::string& attrName, bool f);
	LightAttributeSet<number> const& getNumberAttrs() const;
	LightAttributeSet<std::string> const& getStringAttrs() const;
	LightAttributeSet<bool> const& getBoolAttrs() const;
	AttributeSet<LightEntity*> const& getEntityAttrs() const;
	bool isEmpty() const;
	template<typename T> T& get(std::string const& attr);
	template<typename T> T const& get(std::string const& attr) const;
	bool hasAttributeOfName(std::string const& attrName) const;
	bool isStatic(std::string const& attrName) const;
	bool isDynamic(std::string const& attrName) const;
	bool isAtom(std::string const& attrName) const;
	bool isSet(std::string const& attrName) const;
	std::list<std::string> getAllAttributeNames() const;
	std::string toString() const;
	std::string operator[](std::string const& attrName) const;
	std::set<std::string> operator()(std::string const& attrName) const;
	void updateAttrs(std::map<LightEntity*, LightEntity*> const& oldToNewValues);
	template<typename T> bool operator==(T const& other) const;
	template<typename T> bool operator!=(T const& other) const;
	template<typename T> bool operator>>(std::set<T> const& other) const;
	template<typename T> bool operator^=(std::set<T> const& other) const;
	~LightEntity();
};

/** Constructor: creates an entity with EMPTY_ENTITY as name and type.
 * For every attribute list, the default value is given :
 * - number = 0.0
 * - string = ""
 * - bool = false
 * - LightEntity = NULL
 */
inline LightEntity::LightEntity() : name(EMPTY_ENTITY), type(EMPTY_ENTITY),
	numberAttrs(DEFAULT_NUMBER), stringAttrs(DEFAULT_STRING), boolAttrs(DEFAULT_BOOL), entityAttrs(DEFAULT_ENTITY),
	unknown(false), known(false) {}

/** Explicit constructor with just the name.
 * See LightEntity()
 * @param n Entity's name
 */
inline LightEntity::LightEntity(std::string n) : name(n), type(EMPTY_ENTITY),
	numberAttrs(DEFAULT_NUMBER), stringAttrs(DEFAULT_STRING), boolAttrs(DEFAULT_BOOL), entityAttrs(DEFAULT_ENTITY),
	unknown(false), known(false) {}

/** COnstructor from an Entity.
 * The constructor copies dynamic attributes and gets a pointer on statics.
 * @param e Entity to copy
 */
inline LightEntity::LightEntity(Entity const& e) : name(e.getName()), type(e.getType()),
numberAttrs(e.getNumberAttrs()), stringAttrs(e.getStringAttrs()),
boolAttrs(e.getBoolAttrs()), entityAttrs(AttributeSet<LightEntity*>(DEFAULT_ENTITY)), 
unknown(e.isUnknown()), known(e.isKnown()), unknownAttrs(e.getUnknownAttrs()), knownAttrs(e.getKnownAttrs()) {
	AttributeSet<Entity*> eEnt = e.getEntityAttrs();
	std::map<std::string, Entity*> mapAtoms;
	std::map<std::string, Entity*>::const_iterator itAtoms;
	std::map<std::string, std::set<Entity*> > mapSets;
	std::map<std::string, std::set<Entity*> >::const_iterator itSets;
	std::vector<std::string> argStaticAtoms, argDynamicAtoms, argStaticSets, argDynamicSets;
	mapAtoms = eEnt.getStaticAtomAttrs();
	for(itAtoms = mapAtoms.begin() ; itAtoms != mapAtoms.end() ; ++itAtoms)
		argStaticAtoms.push_back(itAtoms->first);
	mapAtoms = eEnt.getDynamicAtomAttrs();
	for(itAtoms = mapAtoms.begin() ; itAtoms != mapAtoms.end() ; ++itAtoms)
		argDynamicAtoms.push_back(itAtoms->first);
	mapSets = eEnt.getStaticSetAttrs();
	for(itSets = mapSets.begin() ; itSets != mapSets.end() ; ++itSets)
		argStaticSets.push_back(itSets->first);
	mapSets = eEnt.getDynamicSetAttrs();
	for(itSets = mapSets.begin() ; itSets != mapSets.end() ; ++itSets)
		argDynamicSets.push_back(itSets->first);
	entityAttrs = AttributeSet<LightEntity*>(DEFAULT_ENTITY, argStaticAtoms, argDynamicAtoms, argStaticSets, argDynamicSets);
}

/** Copy constructor.
 * @param e LightEntity to copy
 */
inline LightEntity::LightEntity(LightEntity const& e) : name(e.getName()), type(e.getType()), numberAttrs(e.getNumberAttrs()), 
	stringAttrs(e.getStringAttrs()), boolAttrs(e.getBoolAttrs()), entityAttrs(e.getEntityAttrs()), unknown(e.isUnknown()), 
	known(e.isKnown()), unknownAttrs(e.getUnknownAttrs()), knownAttrs(e.getKnownAttrs()) {}

/** Get the name.
 * @return name
 */
inline std::string const& LightEntity::getName() const
{
	return name;
}

/** Get the type.
 * @return type
 */
inline std::string const& LightEntity::getType() const
{
	return type;
}

/** Tell if the lightentity is unknown.
 * See Entity::unknown.
 * @return True if the entity is unknown
 */
inline bool LightEntity::isUnknown() const {
	return unknown;
}

/** Tell if the lightentity is known.
 * See Entity::known.
 * @return True if the entity is known
 */
inline bool LightEntity::isKnown() const {
	return known;
}

/** Tell if the attribute is unknown.
 * See LightEntity::unknownAttrs.
 * @param attrName Attribute to check
 * @return True if entity attribute is unknown
 */
inline bool LightEntity::isAttrUnknown(const std::string& attrName) const {
	std::set<std::string>::const_iterator it = unknownAttrs.find(attrName);
	return (it != unknownAttrs.end());
}

/** Tell if the attribute is known.
 * See LightEntity::knownAttrs.
 * @param attrName Attribute to check
 * @return True if entity attribute is known
 */
inline bool LightEntity::isAttrKnown(const std::string& attrName) const {
	std::set<std::string>::const_iterator it = knownAttrs.find(attrName);
	return (it != knownAttrs.end());
}

/** Get the list of unknown attributes.
 * See LightEntity::unknownAttrs.
 * @return List of unknown attributes
 */
inline std::set<std::string> const& LightEntity::getUnknownAttrs() const {
	return unknownAttrs;
}

/** Get the list of known attributes.
 * See LightEntity::knownAttrs.
 * @return List of known attributes
 */
inline std::set<std::string> const& LightEntity::getKnownAttrs() const {
	return knownAttrs;
}

/** Get all the number attributes.
 * @return List of number attributes
 */
inline LightAttributeSet<number> const& LightEntity::getNumberAttrs() const {
	return numberAttrs;
}

/** Get all the string attributes.
 * @return List of string attributes
 */
inline LightAttributeSet<std::string> const& LightEntity::getStringAttrs() const {
	return stringAttrs;
}

/** Get all the boolean attributes.
 * @return List of boolean attributes
 */
inline LightAttributeSet<bool> const& LightEntity::getBoolAttrs() const {
	return boolAttrs;
}

/** Get all the entity attributes.
 * @return List of entity attributes
 */
inline AttributeSet<LightEntity*> const& LightEntity::getEntityAttrs() const {
	return entityAttrs;
}

/** Check if the lightentity is declared empty.
 * The entity is empty if the constructor with no argument was called. In this case both name and type are equal to EMPTY_ENTITY
 * @return True is the entity is declared empty
 */
inline bool LightEntity::isEmpty() const {
	return name == EMPTY_ENTITY && type == EMPTY_ENTITY;
}

//Class only used to trigger the error in LightEntity::get for types that are not supported
//template<typename T>
//struct always_false{
//	enum{ value=false };
//};
// ! Defined in Entity.hh !

/** Get a reference to the value of the attribute named attr.
 * /!\ Triggers an static assertion since it is not allowed to instantiate this get<T>.
 * @param attr Attribute name to get the reference
 * @return Reference to the value of the attribute
 */
template<typename T> inline T& LightEntity::get(std::string const& attr) {
	static_assert(always_false<T>::value, "LightEntity::get does not work with the provided type.");
}

/** Get a const reference to the value of the attribute named attr.
 * /!\ Triggers an static assertion since it is not allowed to instantiate this get<T>.
 * @param attr Attribute name to get the reference
 * @return Const reference to the value of the attribute
 */
template<typename T> inline T const& LightEntity::get(std::string const& attr) const {
	static_assert(always_false<T>::value, "LightEntity::get does not work with the provided type.");
}

/** Get a reference to the value of the number attribute named attr.
 * @param attr Number attribute name to get the reference
 * @return Reference to the value of the attribute
 */
template<> inline number& LightEntity::get<number>(std::string const& attr) {
	try{ return numberAttrs.getAtom(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return numberAttrs.getDefaultAtomValue();
	}
}

/** Get a reference to the value of the number-set attribute named attr.
 * @param attr Number-set attribute name to get the reference
 * @return Reference to the set of value for the attribute
 */
template<> inline std::set<number>& LightEntity::get<std::set<number> >(std::string const& attr) {
	try{ return numberAttrs.getSet(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return numberAttrs.getDefaultSetValue();
	}
}

/** Get a const reference to the value of the number attribute named attr.
 * @param attr Number attribute name to get the reference
 * @return Const reference to the value of the attribute
 */
template<> inline number const& LightEntity::get<number>(std::string const& attr) const {
	try{ return numberAttrs.getAtom(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return numberAttrs.getDefaultAtomValue();
	}
}

/** Get a const reference to the value of the number-set attribute named attr.
 * @param attr Number-set attribute name to get the reference
 * @return Const reference to the set of value for the attribute
 */
template<> inline std::set<number> const& LightEntity::get<std::set<number> >(std::string const& attr) const {
	try{ return numberAttrs.getSet(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return numberAttrs.getDefaultSetValue();
	}
}

/** Get a reference to the value of the string attribute named attr.
 * @param attr String attribute name to get the reference
 * @return Reference to the value of the attribute
 */
template<> inline std::string& LightEntity::get<std::string>(std::string const& attr) {
	try{ return stringAttrs.getAtom(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value returned." << std::endl;
		return stringAttrs.getDefaultAtomValue();
	}
}

/** Get a reference to the value of the string-set attribute named attr.
 * @param attr String-set attribute name to get the reference
 * @return Reference to the set of value for the attribute
 */
template<> inline std::set<std::string>& LightEntity::get<std::set<std::string> >(std::string const& attr) {
	try{ return stringAttrs.getSet(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return stringAttrs.getDefaultSetValue();
	}
}

/** Get a const reference to the value of the string attribute named attr.
 * @param attr String attribute name to get the reference
 * @return Const reference to the value of the attribute
 */
template<> inline std::string const& LightEntity::get<std::string>(std::string const& attr) const {
	try{ return stringAttrs.getAtom(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return stringAttrs.getDefaultAtomValue();
	}
}

/** Get a const reference to the value of the string-set attribute named attr.
 * @param attr String-set attribute name to get the reference
 * @return Cosnt reference to the set of value for the attribute
 */
template<> inline std::set<std::string> const& LightEntity::get<std::set<std::string> >(std::string const& attr) const {
	try{ return stringAttrs.getSet(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return stringAttrs.getDefaultSetValue();
	}
}

/** Get a reference to the value of the boolean attribute named attr.
 * @param attr Boolean attribute name to get the reference
 * @return Reference to the value of the attribute
 */
template<> inline bool& LightEntity::get<bool>(std::string const& attr) {
	try{ return boolAttrs.getAtom(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return boolAttrs.getDefaultAtomValue();
	}
}

/** Get a reference to the value of the boolean-set attribute named attr.
 * @param attr Boolean-set attribute name to get the reference
 * @return Reference to the set of value for the attribute
 */
template<> inline std::set<bool>& LightEntity::get<std::set<bool> >(std::string const& attr) {
	try{ return boolAttrs.getSet(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return boolAttrs.getDefaultSetValue();
	}
}

/** Get a const reference to the value of the boolean attribute named attr.
 * @param attr Boolean attribute name to get the reference
 * @return Const reference to the value of the attribute
 */
template<> inline bool const& LightEntity::get<bool>(std::string const& attr) const {
	try{ return boolAttrs.getAtom(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return boolAttrs.getDefaultAtomValue();
	}
}

/** Get a const reference to the value of the boolean-set attribute named attr.
 * @param attr Boolean-set attribute name to get the reference
 * @return Const reference to the set of value for the attribute
 */
template<> inline std::set<bool> const& LightEntity::get<std::set<bool> >(std::string const& attr) const {
	try{ return boolAttrs.getSet(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return boolAttrs.getDefaultSetValue();
	}
}

/** Get a reference to the value of the lightentity attribute named attr.
 * @param attr LightEntity attribute name to get the reference
 * @return Reference to the value of the attribute
 */
template<> inline LightEntity*& LightEntity::get<LightEntity*>(std::string const& attr) {
	try{ return entityAttrs.getAtom(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return entityAttrs.getDefaultAtomValue();
	}
}

/** Get a reference to the value of the lightentity-set attribute named attr.
 * @param attr LightEntity-set attribute name to get the reference
 * @return Reference to the set of value for the attribute
 */
template<> inline std::set<LightEntity*>& LightEntity::get<std::set<LightEntity*> >(std::string const& attr) {
	try{ return entityAttrs.getSet(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return entityAttrs.getDefaultSetValue();
	}
}

/** Get a const reference to the value of the lightentity attribute named attr.
 * @param attr LightEntity attribute name to get the reference
 * @return Const reference to the value of the attribute
 */
template<> inline LightEntity* const& LightEntity::get<LightEntity*>(std::string const& attr) const {
	try{ return entityAttrs.getAtom(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return entityAttrs.getDefaultAtomValue();
	}
}

/** Get a const reference to the value of the lightentity-set attribute named attr.
 * @param attr LightEntity-set attribute name to get the reference
 * @return Const reference to the set of value for the attribute
 */
template<> inline std::set<LightEntity*> const& LightEntity::get<std::set<LightEntity*> >(std::string const& attr) const {
	try{ return entityAttrs.getSet(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return entityAttrs.getDefaultSetValue();
	}
}

/** Update the lightentity's attributes.
 * See AttributeSet::updateAttrs
 * @param oldToNewValues A map contaning as key (first row) the previous values to change, and as content (second row) the new values to replace with.
 */
inline void LightEntity::updateAttrs(std::map<LightEntity*, LightEntity*> const& oldToNewValues) {
	entityAttrs.updateAttrs(oldToNewValues);
}

/** Make a lexical cast on "other" to convert it to a string and make a comparison to this->name.
 * @param other Object to compare to
 * @return True if match
 */
template<typename T> inline bool LightEntity::operator==(T const& other) const {
	try {
		std::string sOther = boost::lexical_cast<std::string>(other);
		return name == sOther;
	}
	catch(boost::bad_lexical_cast &) {
		std::cerr << "ERROR: trying to compare an Entity with a type non-convertible in std::string with ==, ";
		std::cerr << "false returned." << std::endl;
		return false;
	}
}

/** Compare this lightentity with the given entity using their name.
 * @param e Entity to compare this to
 * @return True if the two names match
 */
template<> inline bool LightEntity::operator==(Entity const& e) const {
	return name == e.getName();
}

/** Make a lexical cast on "other" to convert it to a string and compare it with this->name to check if different.
 * @param other Object to compare with
 * @return True if different
 */
template<typename T> inline bool LightEntity::operator!=(T const& other) const {
	try {
		std::string sOther = boost::lexical_cast<std::string>(other);
		return name != sOther;
	}
	catch(boost::bad_lexical_cast &) {
		std::cerr << "ERROR: trying to compare an Entity with a type non-convertible in std::string with !=, ";
		std::cerr << "false returned." << std::endl;
		return false;
	}
}

/** Compare this lightentity with the given entity using their name.
 * @param e Entity to compare this to
 * @return True if the two names are differents
 */
template<> inline bool LightEntity::operator!=(Entity const& e) const
{
	return name != e.getName();
}

/** "Is in" operator. Check if this is in the given set.
 * The objects stored are cast lexically and compared with this->name.
 * @param other Set to inspect
 * @return True if this (actually this->name) is in the set
 */
template<typename T> inline bool LightEntity::operator>>(std::set<T> const& other) const {
	typename std::set<T>::const_iterator it;
	try {
		std::string sOther;
		for(it=other.begin() ; it != other.end() ; ++it) {
			sOther = boost::lexical_cast<std::string>(other);
			if(name == sOther) return true;
		}
		return false;
	}
	catch(boost::bad_lexical_cast &) {
		std::cerr << "ERROR: trying to compare an Entity with a type non-convertible in std::string with >>, ";
		std::cerr << "false returned." << std::endl;
		return false;
	}
}

/** "Is in" operator. Check if this is in the given set.
 * The entities are compared using their name.
 * @param s Set to inspect
 * @return True if this (actually this->name) is in the set
 */
template<> inline bool LightEntity::operator>>(std::set<Entity*> const& s) const {
	for(std::set<Entity*>::const_iterator it=s.begin() ; it != s.end() ; ++it)
		if(name == (*it)->getName()) return true;
	return false;
}

/** "Is not in" operator. Check if this is in the given set.
 * See LightEntity::operator>>(std::set<T> const& other) const as it is the opposit.
 * @param other Set to inspect
 * @return True if this->name is not in the set
 */
template<typename T> inline bool LightEntity::operator^=(std::set<T> const& other) const {
	return !(*this >> other);
}

/** "Is not in" operator. Check if this is in the given Entity set.
 * See LightEntity::operator>>(std::set<T> const& other) const as it is the opposit.
 * @param s Set to inspect
 * @return True if this->name is not in the set
 */
template<> inline bool LightEntity::operator^=(std::set<Entity*> const& s) const {
	for(std::set<Entity*>::const_iterator it=s.begin() ; it != s.end() ; ++it)
		if(name == (*it)->getName()) return false;
	return true;
}

/** Destructor: empty
 */
inline LightEntity::~LightEntity()
{}

/** Specialised version of toStringStream for entity attributes.
 * See AttributeSet<T>::toStringStream(std::ostringstream& st)
 * @param st The stream the attributes will be printed in
 */
template<> inline void AttributeSet<LightEntity*>::toStringStream(std::ostringstream& st) const {
	std::map<std::string, LightEntity*>::const_iterator it1;
	std::map<std::string, std::set<LightEntity*> >::const_iterator it2;
	std::set<LightEntity*>::const_iterator setIT;
	std::set<LightEntity*> setData;
	std::string sNone("<none>");
	for(it1=staticAtomAttrs.begin() ; it1 != staticAtomAttrs.end() ; ++it1) {
		st << "\t\t" << it1->first + ": ";
		if(it1->second) st << it1->second->getName() << "(" << it1->second << " )";
		else st << sNone;
		st << " (static)" << std::endl;
	}
	
	for(it1=dynamicAtomAttrs.begin() ; it1 != dynamicAtomAttrs.end() ; ++it1) {
		st << "\t\t" << it1->first + ": ";
		if(it1->second) st << it1->second->getName() << "(" << it1->second << ")";
		else st << sNone;
		st << " (dynamic)" << std::endl;
	}
	for(it2=staticSetAttrs.begin() ; it2 != staticSetAttrs.end() ; ++it2) {
		setData = it2->second;
		st << "\t\t" << it2->first << ": {";
		for(setIT=setData.begin() ; setIT != setData.end() ;) {
	 		st << (*setIT)->getName() << "(" << *setIT << ")";
			if(++setIT != setData.end()) st << ", ";
	 	}
		st << "} (static)" << std::endl;
	}
	for(it2=dynamicSetAttrs.begin() ; it2 != dynamicSetAttrs.end() ; ++it2) {
		setData = it2->second;
		st << "\t\t" << it2->first << ": {";
		for(setIT=setData.begin() ; setIT != setData.end() ;) {
			st << (*setIT)->getName() << "(" << *setIT << ")";
			if(++setIT != setData.end()) st << ", ";
	 	}
		st << "} (dynamic)" << std::endl;
	}
}

#endif
