#include<iostream>

#include "WorldBase.hh"
#include "tools.hh"

using namespace std;

/** Add an Entity to the database.
 * If there already is an Entity with the same name it gives a warning and replace the old Entity (whitout freeing the last).
 * @param e Entity to add to the database
 */
void WorldBase::add(Entity const& e) {
	if(database.find(e.getName()) == database.end()) database[e.getName()] = new Entity(e);
	else cerr << "WARNING: trying to add entity \"" << e.getName() << "\" whereas it already exists" << endl;
}

/** Clear all dynamic vectors of all the entities in the database.
 * It means that all dynamic attributes (of every entity) are set to their default value. (See AttributeSet::delallDynamicVector())
 */
void WorldBase::clearAllDynamicVectors() {
	for(map<string, Entity*>::iterator it=database.begin() ; it != database.end() ; ++it)
		(it->second)->delallDynamicVector();
}


/** Check if an Entity exists in the agent's database.
 * @param sName Name of the entity to check
 * @return True if the entity exists
 */
bool WorldBase::containEntity(std::string const& sName) const {
	return database.find(sName) != database.end();
}

/** Check if an Entity (with a specific type) exists in the agent's database.
 * @param sName Name of the entity to check
 * @param sType The type the entity need to be
 * @return True if the entity exists and is of the given type
 */
bool WorldBase::containEntity(std::string const& sName, std::string const& sType) const {
	map<string, Entity*>::const_iterator it = database.find(sName);
	if(it != database.end() && it->second->getType() == sType)
		return true;

	return false;
}

/** Find an Entity in the world database using its name and type.
 * @param sName The entity to find
 * @param sType The type of the entity
 * @return A pointer to the entity or NULL if the entity does not exist or is not of the given type
 */
Entity* WorldBase::find(std::string const& sName, std::string const& sType) {
	map<string, Entity*>::iterator it = database.find(sName);
	if(it != database.end() && it->second->getType() == sType)
		return it->second;
	cerr << "ERROR when trying to access to an unknown Entity \"" << sName << "\" of type \"" << sType;
	cerr << "\" in the WorldBase." << endl << "NULL returned instead" << endl;
	return NULL;
}

/** Find an entity in the world database using its name and type.
 * @param sName The entity to find
 * @param sType The type of the entity
 * @return A const pointer to the entity or NULL if the entity does not exist or is not of the given type
 */
Entity const* WorldBase::find(std::string const& sName, std::string const& sType) const {
	map<string, Entity*>::const_iterator it = database.find(sName);
	if(it != database.end() && it->second->getType() == sType)
		return it->second;
	cerr << "ERROR when trying to access to an unknown Entity \"" << sName << "\" of type \"" << sType;
	cerr << "\" in the WorldBase." << endl << "NULL returned instead" << endl;
	return NULL;
}

/** Find a set of all entities matching the given type.
 * @param sType The type the entities need to match
 * @return Set of entities matching the given type
 */
vector<Entity*> WorldBase::findByType(string const& sType) const {
	vector<Entity*> vec;
	for(map<string,Entity*>::const_iterator it=database.begin() ; it!=database.end() ; ++it)
		if(it->second->getType() == sType)
			vec.push_back(it->second);
	return vec;
}

/** Find a set of all entities that have a given attribute.
 * @param attrName The attribute the entities need to have
 * @return Set of entities having the given attribute
 */
vector<Entity*> WorldBase::findByAttribute(string const& attrName) const {
	vector<Entity*> vec;
	for(map<string,Entity*>::const_iterator it=database.begin() ; it!=database.end() ; ++it) {
		if(it->second->hasAttributeOfName(attrName)) {
			vec.push_back(it->second);
		}
	}
	return vec;
}

/** Set the name of this database.
 * @param agentName New name fot this database
 */
void WorldBase::setMyselfName(std::string const& agentName) {
	this->myselfName = agentName;
	this->activeBM = true;
}

/** Create a WorldBaseAgent for a given agent.
 * If the agent already had a database it gives a warning and does not replace the previous.
 * @param agentName The name of the agent whose base has to be initialised
 */
void WorldBase::createWBForAgent(string const& agentName) {
	if(this->WBAgents.find(agentName) == this->WBAgents.end()) 
		this->WBAgents[agentName] = new WorldBaseAgent();
	else cerr << "WARNING: trying to add Agent \"" << agentName << "\" DB whereas it already exists" << endl;
}

/** Return true if the given agent exists
 * @param agentName Agent to check
 * @return True if the agent exists
 */
bool WorldBase::hasAgent(std::string const& agentName){
	return(this->WBAgents.find(agentName) != this->WBAgents.end());
}

/** Add an entity to a given agent's database.
 * @param agentName Agent to add the entity to
 * @param e Entity to add
 */
void WorldBase::addForAgent(string const& agentName, Entity const& e) {
	map<string, WorldBaseAgent*>::iterator it = WBAgents.find(agentName);
	if(it != WBAgents.end()) it->second->add(e);
	else cerr << "WARNING: trying to add \"" << e.getName() << "\" in an unexisting agent database" << endl;
}

/** Try to add an entity to an agent's database.
 * If the agent does not already have a database, it is created. If the entity already exists in the agent's database then a pointer to this entiy is returned. If the entity does not exist, it is copied from this database (if possible).
 * @param agentName Name of the agent to add the entity to
 * @param entityName Name of the entity to add
 * @return Pointer to this entity or NULL in case of error (the agent or the entity does not exist)
 */
Entity* WorldBase::initForAgent(std::string const& agentName, std::string const& entityName) {
	if(this->WBAgents.find(agentName) == this->WBAgents.end()) {
		this->createWBForAgent(agentName);
	}
	map<string, WorldBaseAgent*>::iterator it = WBAgents.find(agentName);
	if(it != WBAgents.end()) {
		Entity* tmp = it->second->find(entityName);
		if(tmp != NULL) {
			return tmp;
		}
		else {
			tmp = this->find(entityName);
			if(tmp == NULL) return NULL;
			Entity* newEnt = new Entity(*tmp);
			newEnt->initAttrsKnownUnknown();
			newEnt->setKnown(false);
			newEnt->setUnknown(false);
			it->second->add(*newEnt);
			return it->second->find(entityName);
		}
	}
	return NULL;
}

/** Update an attribute for an agent's entity using the value from another agent.
 * @param fromAgent Agent from which to copy the value
 * @param toAgent Agent that has the entity to be updated
 * @param entity Name of the entity to update
 * @param attribute Name of the entity's attribute to update
 */
void WorldBase::updateForAgent(std::string const& fromAgent, std::string const& toAgent, std::string const& entity, std::string const& attribute) {
	Entity* eFromAgent = this->findForAgent(fromAgent, entity);
	Entity* eToAgent = this->findForAgent(toAgent, entity);
	if(attribute == string("NULL")) {
		if(eToAgent!=NULL) delete(eToAgent);
		eToAgent = new Entity(*eFromAgent);
	}
	else {
		if(eFromAgent != NULL && eToAgent != NULL) {
			if(eFromAgent->isAtom(attribute)) {
				if(eFromAgent->getNumberAttrs().hasAttributeOfName(attribute)) eToAgent->get<number>(attribute) = eFromAgent->get<number>(attribute);
				if(eFromAgent->getStringAttrs().hasAttributeOfName(attribute)) eToAgent->get<string>(attribute) = eFromAgent->get<string>(attribute);
				if(eFromAgent->getBoolAttrs().hasAttributeOfName(attribute)) eToAgent->get<bool>(attribute) = eFromAgent->get<bool>(attribute);
				if(eFromAgent->getEntityAttrs().hasAttributeOfName(attribute)) eToAgent->get<Entity*>(attribute) = eFromAgent->get<Entity*>(attribute);
			}
			else {
				if(eFromAgent->getNumberAttrs().hasAttributeOfName(attribute)) eToAgent->get<set<number> >(attribute) = eFromAgent->get<set<number> >(attribute);
				if(eFromAgent->getStringAttrs().hasAttributeOfName(attribute)) eToAgent->get<set<string> >(attribute) = eFromAgent->get<set<string> >(attribute);
				if(eFromAgent->getBoolAttrs().hasAttributeOfName(attribute)) eToAgent->get<set<bool> >(attribute) = eFromAgent->get<set<bool> >(attribute);
				if(eFromAgent->getEntityAttrs().hasAttributeOfName(attribute)) eToAgent->get<set<Entity*> >(attribute) = eFromAgent->get<set<Entity*> >(attribute);
			}
			eToAgent->setAttrKnown(attribute, false);
			eToAgent->setAttrUnknown(attribute, false);
		}
		else {
			cerr << "ERROR in UpdateForAgent. Something is missing in DB." << endl;
		}
	}
}

Entity* WorldBase::getEntity(std::string const& agentName, std::string const& entityName){
	if(isActiveBM() && !hasAgent(agentName)){
		cerr << "ERROR the request agent " << agentName << " does not exist in the domain" << endl;
		return NULL;
	}
	
	Entity* entity;
	
	if(isActiveBM()){
		entity=findForAgent(agentName, entityName);
		if(entity==NULL){
			cerr << "ERROR in getEntity, the entity " << entityName;
			cerr << " does not exist for agent " << agentName << std::endl;
			return NULL;
		}
	}else{
		if(!containEntity(entityName)){
			cerr << "ERROR in getEntity, the entity " << entityName << " does not exist" << std::endl;
			return NULL;
		}
		entity=find(entityName);
	}
	
	return entity;
}

bool WorldBase::updateAtomAttribute(std::string const& agentName, std::string const& entityName, std::string const& attrName, std::string const& value){
	Entity* entity=getEntity(agentName, entityName);

	//From here entity contains the entity (whether coming from database or agent's database)
	if(!entity->hasAttributeOfName(attrName) || entity->isSet(attrName)){
		cerr << "ERROR: trying to modify an unknown attribute " << attrName;
		cerr << " of entity" << entityName << endl;
		return false;
	}
	
	//Handle belief management, if the value is not related to BM then it jumps after the else
	if(isActiveBM()){
		if(value=="known"){
			entity->setAttrKnown(attrName, true);
			return true;
		}
		
		if(value=="unknown"){
			entity->setAttrUnknown(attrName, true);
			return true;
		}
	}

	if(entity->getNumberAttrs().hasAttributeOfName(attrName)){
		try {
			number temp = boost::lexical_cast<number>(value);
			entity->get<number>(attrName)=temp;
		}
		catch(boost::bad_lexical_cast &) {
			std::cerr << "ERROR: trying to put a value (not of type number) in the attribute "<<attrName;
			std::cerr << ", false returned." << std::endl;
			return false;
		}
		return true;
	}
	if(entity->getStringAttrs().hasAttributeOfName(attrName)){
		entity->get<string>(attrName)=value;
		return true;
	}
	if(entity->getBoolAttrs().hasAttributeOfName(attrName)){
		try {
			bool temp = boost::lexical_cast<bool>(value);
			entity->get<bool>(attrName)=temp;
		}
		catch(boost::bad_lexical_cast &) {
			std::cerr << "ERROR: trying to put a value (not of type bool) in the attribute "<<attrName;
			std::cerr << ", false returned." << std::endl;
			return false;
		}
		return true;
	}
	if(entity->getEntityAttrs().hasAttributeOfName(attrName)){
		if(value!="NULL"){
			Entity* temp=find(value);
			if(temp==NULL){
				std::cerr << "ERROR: trying to put a value (not of type Entity) in the attribute "<<attrName;
				std::cerr << ", false returned." << std::endl;
				return false;
			}
			entity->get<Entity*>(attrName)=temp;
		}else{
			entity->get<Entity*>(attrName)=NULL;
		}
		return true;
	}

	return false;
}

bool WorldBase::clearSetAttribute(std::string const& agentName, std::string const& entityName, std::string const& attrName){
	Entity* entity=getEntity(agentName, entityName);
	if(entity==NULL) return false;
	
	if( !(entity->hasAttributeOfName(attrName)) || entity->isAtom(attrName) ) {
		cerr << "WARNING: trying to modify an unknown attribute \"" << attrName;
		cerr << "\", false returned" << endl;
		return false;
	}
	
	if(entity->getNumberAttrs().hasAttributeOfName(attrName)){
		entity->get< std::set<number> >(attrName).clear();
		return true;
	}
	if(entity->getStringAttrs().hasAttributeOfName(attrName)){
		entity->get< std::set<std::string> >(attrName).clear();
		return true;
	}
	if(entity->getBoolAttrs().hasAttributeOfName(attrName)){
		entity->get< std::set<bool> >(attrName).clear();
		return true;
	}
	if(entity->getEntityAttrs().hasAttributeOfName(attrName)){
		entity->get< std::set<Entity*> >(attrName).clear();
		return true;
	}
	return false;
}

bool WorldBase::addToSetAttribute(std::string const& agentName, std::string const& entityName, std::string const& attrName, std::string const& value){
	Entity* entity=getEntity(agentName, entityName);
	if(entity==NULL) return false;
	
	if( !(entity->hasAttributeOfName(attrName)) || entity->isAtom(attrName) ) {
		cerr << "WARNING: trying to modify an unknown attribute \"" << attrName;
		cerr << "\", false returned" << endl;
		return false;
	}
	
	if(entity->getNumberAttrs().hasAttributeOfName(attrName)){
		try {
			number temp = boost::lexical_cast<number>(value);
			entity->get< std::set<number> >(attrName).insert(temp);
		}
		catch(boost::bad_lexical_cast &) {
			std::cerr << "ERROR: trying to put a value (not of type number) in the attribute "<<attrName;
			std::cerr << ", false returned." << std::endl;
			return false;
		}
		return true;
	}
	if(entity->getStringAttrs().hasAttributeOfName(attrName)){
		entity->get< std::set<string> >(attrName).insert(value);
		return true;
	}
	if(entity->getBoolAttrs().hasAttributeOfName(attrName)){
		try {
			bool temp = boost::lexical_cast<bool>(value);
			entity->get< std::set<bool> >(attrName).insert(temp);
		}
		catch(boost::bad_lexical_cast &) {
			std::cerr << "ERROR: trying to put a value (not of type bool) in the attribute "<<attrName;
			std::cerr << ", false returned." << std::endl;
			return false;
		}
		return true;
	}
	if(entity->getEntityAttrs().hasAttributeOfName(attrName)){
		Entity* temp=find(value);
		if(temp==NULL){
			std::cerr << "ERROR: trying to put a value (not of type Entity) in the attribute "<<attrName;
			std::cerr << ", false returned." << std::endl;
			return false;
		}
		entity->get< std::set<Entity*> >(attrName).insert(temp);
		return true;
	}
	return false;
}

/** Create a string representing the WorldBase.
 * First put the list of entities in this database, then put every agent database. (See WorldBaseAgent::toString().)
 * @return A string representing the world entity databse
 */
string WorldBase::toString() const {
	string s = "***** Start Myself WorldBase Printing *****\n";
	for(map<string,Entity*>::const_iterator it=database.begin() ; it!=database.end() ; ++it) {
		s += it->second->toString() + "\n";
	}
	s += "***** End Myself WorldBase Printing *****\n";
	for(map<string,WorldBaseAgent*>::const_iterator it2=WBAgents.begin() ; it2!=WBAgents.end() ; ++it2) {
		s += "\nFOR AGENT "+it2->first+" :\n\n";
		s += it2->second->toString() + "\n";
	}
	return s;
}

/** Return a reference to an Entity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param c Entity to return
 * @return Reference to the entity, or to an empty entity
 */
Entity& WorldBase::operator[](const char* c) {
	std::map<std::string, Entity*>::iterator it = database.find(c);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the WorldBase: " << c;
		cerr << " has not been found" << endl << "default Entity returned instead" << endl;
		return defaultEntity;
	}
}

/** Return a reference to an Entity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param s Entity to return
 * @return Reference to the entity, or to an empty entity
 */
Entity& WorldBase::operator[](std::string const& s) {
	std::map<std::string, Entity*>::iterator it = database.find(s);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the WorldBase: " << s;
		cerr << " has not been found" << endl << "default Entity returned instead" << endl;
		return defaultEntity;
	}
}

/** Return a reference to an Entity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param c Entity to return
 * @return Const reference to the entity, or to an empty entity
 */
Entity const& WorldBase::operator[](const char* c) const {
	std::map<std::string, Entity*>::const_iterator it = database.find(c);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the WorldBase: " << c;
		cerr << " has not been found" << endl << "default Entity returned instead" << endl;
		return defaultEntity;
	}
}

/** Return a reference to an Entity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param s Entity to return
 * @return Const reference to the entity, or to an empty entity
 */
Entity const& WorldBase::operator[](std::string const& s) const {
	std::map<std::string, Entity*>::const_iterator it = database.find(s);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the WorldBase: " << s;
		cerr << " has not been found" << endl << "default Entity returned instead" << endl;
		return defaultEntity;
	}
}

/** Destructor : delete all Entities in database
 */
WorldBase::~WorldBase() {
	for(map<string,WorldBaseAgent*>::iterator it=WBAgents.begin() ; it!=WBAgents.end() ; ++it)
		delete it->second;
	for(map<string,Entity*>::iterator it=database.begin() ; it!=database.end() ; ++it)
		delete it->second;
}
